
#include "app.h"
#include "Horde3D.h"
#include "Horde3DUtils.h"
#include <math.h>
#include <iomanip>
#include <windows.h>


char path[MAX_PATH]="";
// Convert from degrees to radians
inline float degToRad( float f ) 
{
	return f * (3.1415926f / 180.0f);
}


Application::Application( const string &contentDir )
{
	for( unsigned int i = 0; i < 320; ++i ) _keys[i] = false;

	_x = 0; _y = 1; _z = 1; _rx = 7; _ry = 15; _velocity = 5.0f;
	_curFPS = 30;

	_freeze = false; _showStats = true; _debugViewMode = false; _wireframeMode = false;
	_animTime = 0; _weight = 1.0f;
	_cam = 0; _light = 0;

	_contentDir = contentDir;
}


bool Application::reload_resources() {

	H3DRes reshandle=0;
	int restype;

	while(reshandle=h3dGetNextResource(H3DResTypes::Undefined,reshandle)) {
		restype=h3dGetResType(reshandle);
		h3dUnloadResource(reshandle);
	}

	h3dutLoadResourcesFromDisk("./content");

	H3DRes matRes = h3dFindResource( H3DResTypes::Material, "pipelines/postHDR.material.xml" );
//	NodeHandle matRes = h3dfindResource( H3DResTypes::Material, "pipelines/postHDR.material.xml" );
	// hdrParams: exposure, brightpass threshold, brightpass offset (see shader for description)
	h3dSetMaterialUniform( matRes, "hdrParams", 10.0f, 0.5f, 0.08f, 0 );
	return true;
}

void Application::dumpMessages() {

	int level;
	float timestamp;
	static char *errors[]={ "ERROR", "WARN", "INFO", "DEBUG" };

	while(1) {
		const char *mes = h3dGetMessage(&level,&timestamp);
		if(*mes==0) break;
		printf("%s: %s\n",errors[level-1],mes);
	}
}

bool Application::init()
{	
	// Initialize engine
	if( !h3dInit() )
	{	
		dumpMessages();
		return false;
	}

	// Set options
	h3dSetOption( H3DOptions::LoadTextures, 1 );
	h3dSetOption( H3DOptions::TexCompression, 0 );
	h3dSetOption( H3DOptions::FastAnimation, 0 );
	h3dSetOption( H3DOptions::MaxAnisotropy, 8 );
	h3dSetOption( H3DOptions::ShadowMapSize, 2048 );

	// Add resources
	// Pipelines
	_hdrPipeRes = h3dAddResource( H3DResTypes::Pipeline, "pipelines/hdr.pipeline.xml", 0 );
	_forwardPipeRes = h3dAddResource( H3DResTypes::Pipeline, "pipelines/forward.pipeline.xml", 0 );
	// Font
	_fontMatRes = h3dAddResource( H3DResTypes::Material, "overlays/font.material.xml", 0 );
	_panelMatRes = h3dAddResource( H3DResTypes::Material, "overlays/panel.material.xml", 0 );

	// scene
	if(!*path) {
		OPENFILENAME ofn;   
		char szFile[260] =""; // Buffer for selected file name.
		char szFileTitle[260] =""; // Buffer for selected file name.
		memset(&ofn,0,sizeof(OPENFILENAME)); 
		ofn.lStructSize         = sizeof(OPENFILENAME); 
		ofn.hwndOwner           = NULL; 
		ofn.lpstrFilter         = "Scene Files (*.xml; *.scn)\0*.xml;*.scn\0\0";; 
		ofn.nFilterIndex        = 1; 
		ofn.lpstrFile           = szFile; 
		ofn.nMaxFile            = MAX_PATH; 
		ofn.lpstrTitle          = "Open Scene File"; 
		ofn.lpstrFileTitle		= szFileTitle;
		ofn.nMaxFileTitle		= MAX_PATH;
		ofn.Flags               = OFN_FILEMUSTEXIST|OFN_NOCHANGEDIR ; 
		ofn.lpstrInitialDir		= "./content";
		if (GetOpenFileName(&ofn) != TRUE) exit(0);

		strcpy(path,strstr(ofn.lpstrFile,"content")+8);
	}
		printf("loading: %s\n",path);
	H3DRes sceneres = h3dAddResource( H3DResTypes::SceneGraph, path, 0 );
	
	// Load resources
	h3dutLoadResourcesFromDisk( _contentDir.c_str() );

	// Add scene nodes
	// Add camera
	_cam = h3dAddCameraNode( H3DRootNode, "Camera", _hdrPipeRes );


	H3DNode scene = h3dAddNodes( H3DRootNode, sceneres );
//	h3dsetNodeTransform( scene, 0, 0, 0, 0, 0, 0, 1.0f, 1.0f, 1.0f );

	// Add scene
//	for(int x=0;x<5;x++){
//		for(int z=0;z<5;z++) {
//			NodeHandle scene = h3daddNodes( RootNode, sceneres );
//			h3dsetNodeTransform( scene, x*15.0f, 0, z*15.0f, 0, 0, 0, 10.0f, 10.0f, 10.0f );
//		}
//	}
	// Add light source
	_light = h3dAddLightNode( H3DRootNode, "Light1", 0, "LIGHTING", "SHADOWMAP" );
	h3dSetNodeTransform( _light, 1, 20, 1, -90, 0, 0, 1, 1, 1 );
	h3dSetNodeParamF( _light, H3DLight::RadiusF,0, 20 );
	h3dSetNodeParamF( _light, H3DLight::FovF,0, 90 );
	h3dSetNodeParamI( _light, H3DLight::ShadowMapCountI, 1 );
	h3dSetNodeParamF( _light, H3DLight::ShadowMapBiasF,0, 0.01f );
//	h3dSetNodeParamf( _light, H3DLight::Col_R, 1.0f );
//	h3dSetNodeParamf( _light, H3DLight::Col_G, 0.8f );
//	h3dSetNodeParamf( _light, H3DLight::Col_B, 0.7f );

	// Customize post processing effects
	H3DRes matRes = h3dFindResource( H3DResTypes::Material, "pipelines/postHDR.material.xml" );
	// hdrParams: exposure, brightpass threshold, brightpass offset (see shader for description)
	h3dSetMaterialUniform( matRes, "hdrParams", 10.0f, 0.5f, 0.08f, 0 );
	
	return true;
}


void Application::mainLoop( float fps )
{
	_curFPS = fps;

	keyHandler();
	
	h3dSetOption( H3DOptions::DebugViewMode, _debugViewMode ? 1.0f : 0.0f );
	h3dSetOption( H3DOptions::WireframeMode, _wireframeMode ? 1.0f : 0.0f );
		
	// Set camera parameters
	h3dSetNodeTransform( _cam, _x, _y, _z, _rx ,_ry, 0, 1, 1, 1 );
	h3dSetNodeTransform( _light, _x, _y, _z, _rx ,_ry, 0, 1, 1, 1 );
	
	if( _showStats )
	{
		h3dutShowFrameStats( _fontMatRes, _panelMatRes, 1 );
	}

	
	// Render scene
	h3dRender( _cam );

	// Remove all overlays
	h3dClearOverlays();

	// Write all mesages to log file
	dumpMessages();
}


void Application::release()
{
	// Release engine
	h3dRelease();
}


void Application::resize( int width, int height )
{
	// Resize viewport
	h3dSetupViewport( 0, 0, width, height, true );
	
	// Set virtual camera parameters
	h3dSetupCameraView( _cam, 45.0f, (float)width / height, 0.01f, 10000.0f );
}


void Application::keyPressEvent( int key )
{
	if( key == 32 )		// Space
		_freeze = !_freeze;

	if( key == 260 )	// F3
	{
		if( h3dGetNodeParamI( _cam, H3DCamera::PipeResI ) == _hdrPipeRes )
			h3dSetNodeParamI( _cam, H3DCamera::PipeResI, _forwardPipeRes );
		else
			h3dSetNodeParamI( _cam, H3DCamera::PipeResI, _hdrPipeRes );
	}
	
	if( key == 262) // F5 
		reload_resources();
		//	h3dsetNodeParami( _cam, CameraNodeParams::OcclusionCulling, !h3dgetNodeParami( _cam, CameraNodeParams::OcclusionCulling) );

	if( key == 264 )	// F7
		_debugViewMode = !_debugViewMode;

	if( key == 265 )	// F8
		_wireframeMode = !_wireframeMode;
	
	if( key == 266 )	// F9
		_showStats = !_showStats;
}


void Application::keyHandler()
{
	float curVel = _velocity / _curFPS;
	
	if( _keys[287] ) curVel *= 5;	// LShift
	
	if( _keys['W'] )
	{
		// Move forward
		_x -= sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		_y -= sinf( -degToRad( _rx ) ) * curVel;
		_z -= cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
	}

	if( _keys['S'] )
	{
		// Move backward
		_x += sinf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
		_y += sinf( -degToRad( _rx ) ) * curVel;
		_z += cosf( degToRad( _ry ) ) * cosf( -degToRad( _rx ) ) * curVel;
	}

	if( _keys['A'] )
	{
		// Strafe left
		_x += sinf( degToRad( _ry - 90) ) * curVel;
		_z += cosf( degToRad( _ry - 90 ) ) * curVel;
	}

	if( _keys['D'] )
	{
		// Strafe right
		_x += sinf( degToRad( _ry + 90 ) ) * curVel;
		_z += cosf( degToRad( _ry + 90 ) ) * curVel;
	}

	if( _keys['1'] )
	{
		// Change blend weight
		_weight += 2 / _curFPS;
		if( _weight > 1 ) _weight = 1;
	}
	
	if( _keys['2'] )
	{
		// Change blend weight
		_weight -= 2 / _curFPS;
		if( _weight < 0 ) _weight = 0;
	}
}


void Application::mouseMoveEvent( float dX, float dY )
{
	// Look left/right
	_ry -= dX / 100 * 30;
	
	// Loop up/down but only in a limited range
	_rx += dY / 100 * 30;
	if( _rx > 90 ) _rx = 90; 
	if( _rx < -90 ) _rx = -90;
}
